## io.rb

###### ログファイルIO ######

require 'pstore'
require 'uri'

module WMMC

	class IO

		def initialize(logdir, logsavenum)
			@logdir = logdir
			@logsavenum = logsavenum
			@db = {}
		end

		def opendb(id)
			fid = URI.encode(id, /[^a-zA-Z0-9]/n)
			@db[id] = PStore.new("#{@logdir}/#{fid}")
		end

		def readlog(id, num)
			return [] if id == '' || !id
			opendb(id) unless @db.key?(id)
			log = []
			@db[id].transaction do |db|
				db.roots.sort.reverse.each_with_index do |time, idx|
					log.push(db[time])
					return log if idx >= num - 1
					return log if idx >= 1000
				end
			end
			log
		end

		def writelog(sys, id, peer_id, peer_nick, data, time)
			return true if id == '' || !id
			log_id = (time.to_f * 1000).to_i
			opendb(id) unless @db.key?(id)
			@db[id].transaction do |db|
				while db.root?(log_id)
					log_id += 1
				end
				db[log_id] = [sys, time.to_f, peer_id.to_s, peer_nick.to_s, data.to_s]

				db.roots.sort.reverse.each_with_index do |key, idx|
					next if idx < @logsavenum
					db.delete(key)
				end
			end
			true
		end

		private :opendb

	end

end
