## main.rb

=begin

WebMMC 0.4.2

Copyright (c) 2006-2007 blueday

This program is free software. You can re-distribute and/or
modify this program under the same terms as Ruby itself,
Ruby Distribute License or GNU General Public License.

=end

require 'thread'
require 'uri'

require 'plugins/wmmc/config'
require 'plugins/wmmc/contactlist'
require 'plugins/wmmc/io'
require 'plugins/wmmc/rhtml'

require 'plugins/wmmc/msnmcstm'
require 'plugins/wmmc/msnmconfig'

module WMMC

	WEBMMC_VERSION = '0.4.2'

	class Main

		def initialize(conf)

			@cf = Config.new(conf)

			# ユーザID/パスワードが設定されていないとエラー
			raise 'Please specify set your Messenger ID/Password.' unless @cf.userid || @cf.userpass

			# 会話ログ保存ディレクトリ設定
			@logdir = nil

			if @cf.logdir
				raise "Directory \"#{@cf.logdir}\" not found." unless FileTest.directory?(@cf.logdir)
				@logdir = @cf.logdir
			else
				@logdir = File.dirname(File.expand_path(__FILE__))
				Dir.chdir(@logdir) do
					raise "Directory \"#{@logdir}/wmmc_log\" making failure." unless Dir.mkdir("wmmc_log", 0700) unless FileTest.directory?("#{@logdir}/wmmc_log")
				end
				@logdir = @logdir + '/wmmc_log'
			end

			encoded_userid = URI.encode(@cf.userid, /[^a-zA-Z0-9]/n)

			Dir.chdir(@logdir) do
				raise "Directory \"#{@logdir}/#{encoded_userid}\" making failure." unless Dir.mkdir(encoded_userid, 0700) unless FileTest.directory?("#{@logdir}/#{encoded_userid}")
			end

			@logdir = "#{@logdir}/#{encoded_userid}"


			@queue = Queue.new
			@method = 'post'
			@connect = false
			@mystatus = 'FLN'
			@msnm = nil
			@msnm_thread = nil
			@msnm_ping_timer = nil
			@log_io = WMMC::IO.new(@logdir, @cf.logsavenum)
			@contactlist = WMMC::ContactList.new(@logdir)
			@rhtml = WMMC::RHTML.new(@method)
			@latest_message = {}
			@session_users = {}


			# MSN Messenger Client（別スレッドで動作）
			@msnm_client = Proc.new do
				@msnm_thread = Thread.new do
					puts 'MSN Messenger Client start.'
					# Webインターフェースからのキューを読んで動作
					while q = @queue.deq
						begin
							case q[0]

							# ログイン
							when 'login'
								if !@connect
									@msnm = MSNMessenger.new(@cf.userid, @cf.userpass, WebMessengerSessionHandlerFactory.new, @log_io)
									@msnm.ns.synchronize
									sleep(3)
									@msnm.set_status(q[1])
									@connect = true
									sleep(3)
									@msnm.myname = @cf.usernick
									@mystatus = @msnm.mystatus
									@msnm.ns.set_nick(@cf.userid, @cf.usernick) if @cf.usernick
									sleep(3)
									@contactlist.update(@msnm.userlist)
								end

							# ログアウト
							when 'logout'
								if @connect
									@contactlist.update(@msnm.userlist)
									@msnm.logout
								end
								@connect = false
								@msnm = nil
								@mystatus = 'FLN'

							# ステータス変更
							when 'changestatus'
								@msnm.set_status(q[1]) if @msnm

							# コンタクトリスト更新
							when 'listupdate'
								@contactlist.update(@msnm.userlist) if @msnm
								@mystatus = @msnm.mystatus if @msnm

							# セッションユーザ取得
							when 'getsessionusers'
								@session_users[q[1]] = @msnm.get_session_users(q[1]) if @msnm

							# 最新発言者リスト取得
							when 'infoupdate'
								@latest_message.update(@msnm.latest_message) if @msnm

							# メッセージ送信
							when 'sendmessage'
								@msnm.send_message_to_peer(q[1], q[2])

							# セッション切断
							when 'disconnectsession'
								@msnm.disconnect_session(q[1])

							# シャットダウン
							when 'shutdown'
								@msnm.logout if @msnm && @connect
								@session_users = {}
								@connect = false
								break
							end

						rescue
							puts 'Thread irregular happened.'
						end
					end
					puts 'MSN Messenger Client stop.'
				end

				@msnm_ping_timer = Thread.new do
					begin
						sleep(60)
						while @connect
							@msnm.send_ping
							sleep(60)
						end
					rescue
					end
				end
			end

		end # def initialize


		def mountproc

			# Webインターフェース表示設定
			Proc.new do |req, res|

				body = ''
				title = ''

				mode = req.query['mode'].to_s
				login = nil
				login = req.query['login'] if req.query['login']
				logout = nil
				logout = req.query['logout'] if req.query['logout']
				message = nil
				message = req.query['message'] unless req.query['message'].to_s == ''
				id = nil
				id = req.query['id'] unless req.query['id'].to_s == ''
				num = 10
				num = req.query['num'].to_i if req.query['num'].to_i > 0
				refresh_on = false
				refresh_on = req.query['refresh_on'] if req.query['refresh_on']
				sendmessage = nil
				sendmessage = req.query['sendmessage'] if req.query['sendmessage']
				status = nil
				status = req.query['status'] unless req.query['status'].to_s == ''
				post = false
				post = true if req.request_method == 'POST'

				@queue.enq(['infoupdate']) if @connect

				@rhtml.set_actionpath(File.basename(req.script_name, '.*'))


				# 入力に応じて動作を設定
				case mode

				# 接続状態変更
				when 'connection'

					# ログイン
					if login && !@connect && post
						@queue.clear
						@msnm_client.call
						@queue.enq(['login', status])
						title = '接続処理中'
						body = @rhtml.connectwaitpage(title)

					# ログアウト
					elsif logout && @connect && post
						@queue.enq(['logout'])
						@contactlist.offline
						@queue.enq(['shutdown'])
						@msnm_ping_timer.exit unless @msnm_ping_timer.join(5)
						unless @msnm_thread.join(5)
							@msnm_thread.exit
							@connect = false
							@msnm = nil
							@mystatus = 'FLN'
						end
						title = '切断'
						body = @rhtml.disconnectwaitpage(title)
					end

				# ステータス変更
				when 'status'
					if status && @connect
						if post
							@queue.enq(['changestatus', status])
							title = 'jump'
							body = @rhtml.jumppage(title, nil, refresh_on)
						end
						@queue.enq(['listupdate'])
					end

				# メッセージ送信
				when 'message'
					if id
						if @connect
							if message && sendmessage && post
								@queue.enq(['sendmessage', message, id]) 
								title = 'jump'
								body = @rhtml.jumppage(title, id, refresh_on)
							end
							@queue.enq(['getsessionusers', id])
							@queue.enq(['listupdate'])
						end
						if body == ''
							title = 'ログ'
							body = @rhtml.logpage(title, @log_io.readlog(id, num), id, @contactlist.get_status(id), @latest_message, @session_users[id], @cf.refresh, refresh_on)
						end
					end

				# ログ表示
				when 'read'
					if id
						if @connect
							@queue.enq(['getsessionusers', id])
							@queue.enq(['listupdate'])
						end
						title = 'ログ'
						body = @rhtml.logpage(title, @log_io.readlog(id, num), id, @contactlist.get_status(id), @latest_message, @session_users[id], @cf.refresh, refresh_on)
					end

				# セッション切断
				when 'disconnect_session'
					if id
						if @connect
							if post
								@queue.enq(['disconnectsession', id])
								title = 'jump'
								body = @rhtml.jumppage(title, id, refresh_on)
							end
							@queue.enq(['getsessionusers', id])
							@queue.enq(['listupdate'])
						end
						if body == ''
							title = 'ログ'
							body = @rhtml.logpage(title, @log_io.readlog(id, num), id, @contactlist.get_status(id), @latest_message, @session_users[id], @cf.refresh, refresh_on)
						end
					end

				# その他
				else
					@queue.enq(['listupdate']) if @connect
				end

				# 動作設定されていなければトップページ表示
				if body == ''
					title = 'トップ'
					body = @rhtml.toppage(title, @cf.userid, @cf.usernick, @mystatus, @connect,  @contactlist.list, @latest_message, WEBMMC_VERSION, @cf.refresh, refresh_on)
				end

				res.content_type = 'text/html; charset=utf-8'
				res.body =  body
			end # mproc

		end # def mountproc


		def startproc
		end


		def stopproc
			if @connect
				@queue.enq(['shutdown'])
				@msnm_ping_timer.exit unless @msnm_ping_timer.join(5)
				@msnm_thread.exit unless @msnm_thread.join(5)
			end
		end

	end

end
